/* Hekr '48-Passthrough' Protocol
 * Copyright(c) 2018 Hangzhou District Nine Technology Co., Ltd.  */

#ifndef HEKR48TRANS_H
#define HEKR48TRANS_H

#include "ra_types.h"

typedef enum
{
    H48_FRAME_TYPE_UNKNOWN = 0xFF,
    H48_FRAME_TYPE_VERIFY_REQUEST_RESP = 0x02,
    H48_FRAME_TYPE_VERIFY_RESP = 0x04,
    H48_FRAME_TYPE_REPORTDEVINFO_RESP = 0x06,
    H48_FRAME_TYPE_APPSEND = 0x07,
    H48_FRAME_TYPE_DEVSEND_RESP = 0x0A,
    H48_FRAME_TYPE_HEARTBEAT_RESP = 0x0C,
} h48_frame_type;


/*******************
 * Utilities
 *******************/


/**
 * Convert raw data to hex-literal
 * @param[out]     dst   buffer for the output data
 * @param[in]      src   input data
 * @param[in]      len   input data length
 * @return Length of the converted data
 * @detail Convert raw data to hex-literal
 */
int raw_to_hexlit(char *dst, const ra_u8 *src, ra_size_t len);

/**
 * Convert hex-literal to raw data
 * @param[out]     dst   buffer for the output data
 * @param[in]      src   input data
 * @param[in]      len   input data length
 * @return Length of the converted data
 * @detail Convert hex-literal to raw data
 */
int hexlit_to_raw(ra_u8 *dst, const char *src, const ra_size_t len);

/**
 * Convert hex-literal to raw data (16-bit)
 * @param[out]     dst   buffer for the output data
 * @param[in]      src   input data
 * @param[in]      len   input data length
 * @return Length of the converted data
 * @detail Convert hex-literal to raw data
 */
int hexlit_to_raw16(ra_u16 *dst, const char *src, const ra_size_t len);



/*******************
 * Frame Building
 *******************/


/**
 * Build frame for verify request (the 1st step of login)
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      buf_size   Buffer size
 * @param[in]      frame_num  Frame number
 * @param[in]      prodkey    Product Key
 * @param[in]      devtid     Device TID
 * @return The length of built frame
 */
int h48_build_frame_verify_request(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, \
        const char *prodkey, const char *devtid);

/**
 * Extract randomkey from verify request response frame
 * @param[out]     buf        Randomkey buffer
 * @param[in]      buf        Randomkey buffer size
 * @param[in]      buf        Verify request response frame data
 * @param[in]      buf        Verify request response frame data size
 * @param[in]      buf_size   Buffer size
 * @return The length of randomkey
 */
int h48_extract_randomkey(char *randomkey_buf, const ra_size_t randomkey_buf_size, \
        ra_u8 *data, const ra_size_t data_size);

/**
 * Build frame for verify (the 2nd step of login)
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      buf_size   Buffer size
 * @param[in]      frame_num  Frame number
 * @param[in]      prodkey    Product Key
 * @param[in]      devtid     Device TID
 * @return The length of built frame
 */
int h48_build_frame_verify(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, \
        const char *randomkey, const char *devtid, const char *devprikey);

/**
 * Build frame for heartbeat
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      buf_size   Buffer size
 * @param[in]      frame_num  Frame number
 * @return The length of built frame
 * @detail Send this frame every 30 seconds
 */
int h48_build_frame_heartbeat(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num);

/**
 * Build frame for data upload from device to cloud
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      buf_size   Buffer size
 * @param[in]      frame_num  Frame number
 * @param[in]      msgid      Message ID
 * @param[in]      data       Data
 * @param[in]      data_len   Data length
 * @return The length of built frame
 * @detail Send data when have some state changes to report
 */
int h48_build_frame_devsend(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, const ra_u16 msgid, \
        const ra_u8 *data, const ra_size_t data_len);

/**
 * Build frame for data upload from device to cloud
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      buf_size   Buffer size
 * @param[in]      frame_num  Frame number
 * @param[in]      msgid      Message ID
 * @param[in]      data       Data
 * @param[in]      data_len   Data length
 * @return The length of built frame
 * @detail Confirm the appsend from cloud
 */
int h48_build_frame_appsendresp(ra_u8 *buf, const ra_size_t buf_size, const ra_u8 frame_num, const ra_u16 msgid, \
        const ra_u8 *apptid);

/**
 * Get the type of a frame
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      frame      Data
 * @param[in]      frame_len  Data length
 * @return The type of frame
 */
h48_frame_type h48_frame_decode_type(ra_u8 *frame, const ra_size_t frame_size);

enum
{
    H48_TRY_SPLIT_FRAME_INCOMPLETE = -1,
    H48_TRY_SPLIT_FRAME_INVALID = -2,
};

/**
 * Test if a frame is complete
 * @param[out]     buf        Buffer for the built frame
 * @param[in]      data       Data
 * @param[in]      data_len   Data length
 * @return         > 0        Frame length (if a complete frame available)
 * @return         -1         Data incomplete
 * @return         -2         Invalid data
 */
int h48_try_split_frame(ra_u8 *data, const ra_size_t data_size);


#endif

